from __future__ import annotations
from typing import Any
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_openpyxl.openpyxl import (Xlsx_readFile_Z721C83C5, Xlsx_load_4E60E31B, Xlsx_read_Z3F6BC7B1, Xlsx_writeFile_A9E7E13, Xlsx_write_Z22AEA2D8)
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from .workbook import (to_fs_workbook, from_fs_workbook)

def _expr390() -> TypeInfo:
    return class_type("FsSpreadsheet.Py.Xlsx", None, Xlsx)


class Xlsx:
    @staticmethod
    def from_xlsx_file(path: str) -> FsWorkbook:
        return to_fs_workbook(Xlsx_readFile_Z721C83C5(path))

    @staticmethod
    def from_xlsx_stream(stream: Any) -> FsWorkbook:
        return to_fs_workbook(Xlsx_load_4E60E31B(stream))

    @staticmethod
    def from_xlsx_bytes(bytes: bytearray) -> FsWorkbook:
        return to_fs_workbook(Xlsx_read_Z3F6BC7B1(bytes))

    @staticmethod
    def to_xlsx_file(path: str, wb: FsWorkbook) -> None:
        Xlsx_writeFile_A9E7E13(from_fs_workbook(wb), path)

    @staticmethod
    def to_xlsx_bytes(wb: FsWorkbook) -> bytearray:
        return Xlsx_write_Z22AEA2D8(from_fs_workbook(wb))


Xlsx_reflection = _expr390

__all__ = ["Xlsx_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_header import (CompositeHeader, IOType)
from ...fable_library.list import (singleton, empty, FSharpList)
from ...fable_library.seq import map
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, IRequiredGetter, string, index, IGetters)
from ...thoth_json_core.encode import list_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)
from .iotype import (encoder as encoder_1, decoder as decoder_1)

__A_ = TypeVar("__A_")

def encoder(ch: CompositeHeader) -> IEncodable:
    def oa_to_json_string(oa: OntologyAnnotation, ch: Any=ch) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    class ObjectExpr2766(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], ch: Any=ch) -> Any:
            return helpers.encode_string(ch.fields[0])

    pattern_input: tuple[str, FSharpList[IEncodable]] = (("Comment", singleton(ObjectExpr2766()))) if (ch.tag == 14) else ((("Parameter", singleton(oa_to_json_string(ch.fields[0])))) if (ch.tag == 3) else ((("Factor", singleton(oa_to_json_string(ch.fields[0])))) if (ch.tag == 2) else ((("Characteristic", singleton(oa_to_json_string(ch.fields[0])))) if (ch.tag == 1) else ((("Component", singleton(oa_to_json_string(ch.fields[0])))) if (ch.tag == 0) else ((("ProtocolType", empty())) if (ch.tag == 4) else ((("ProtocolREF", empty())) if (ch.tag == 8) else ((("ProtocolDescription", empty())) if (ch.tag == 5) else ((("ProtocolUri", empty())) if (ch.tag == 6) else ((("ProtocolVersion", empty())) if (ch.tag == 7) else ((("Performer", empty())) if (ch.tag == 9) else ((("Date", empty())) if (ch.tag == 10) else ((("Input", singleton(encoder_1(ch.fields[0])))) if (ch.tag == 11) else ((("Output", singleton(encoder_1(ch.fields[0])))) if (ch.tag == 12) else ((ch.fields[0], empty())))))))))))))))
    class ObjectExpr2767(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], ch: Any=ch) -> Any:
            return helpers_1.encode_string(pattern_input[0])

    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("headertype", ObjectExpr2767()), ("values", list_1(pattern_input[1]))])
    class ObjectExpr2768(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], ch: Any=ch) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_2))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values_1)
            return helpers_2.encode_object(arg)

    return ObjectExpr2768()


def _arrow2770(get: IGetters) -> CompositeHeader:
    header_type: str
    object_arg: IRequiredGetter = get.Required
    header_type = object_arg.Field("headertype", string)
    def oa(__unit: None=None) -> OntologyAnnotation:
        arg_3: Decoder_1[OntologyAnnotation] = index(0, OntologyAnnotation_decoder)
        object_arg_1: IRequiredGetter = get.Required
        return object_arg_1.Field("values", arg_3)

    def io(__unit: None=None) -> IOType:
        arg_5: Decoder_1[IOType] = index(0, decoder_1)
        object_arg_2: IRequiredGetter = get.Required
        return object_arg_2.Field("values", arg_5)

    def _arrow2769(__unit: None=None) -> str:
        arg_7: Decoder_1[str] = index(0, string)
        object_arg_3: IRequiredGetter = get.Required
        return object_arg_3.Field("values", arg_7)

    return CompositeHeader(1, oa(None)) if (header_type == "Characteristic") else (CompositeHeader(3, oa(None)) if (header_type == "Parameter") else (CompositeHeader(0, oa(None)) if (header_type == "Component") else (CompositeHeader(2, oa(None)) if (header_type == "Factor") else (CompositeHeader(11, io(None)) if (header_type == "Input") else (CompositeHeader(12, io(None)) if (header_type == "Output") else (CompositeHeader(4) if (header_type == "ProtocolType") else (CompositeHeader(8) if (header_type == "ProtocolREF") else (CompositeHeader(5) if (header_type == "ProtocolDescription") else (CompositeHeader(6) if (header_type == "ProtocolUri") else (CompositeHeader(7) if (header_type == "ProtocolVersion") else (CompositeHeader(9) if (header_type == "Performer") else (CompositeHeader(10) if (header_type == "Date") else (CompositeHeader(14, _arrow2769()) if (header_type == "Comment") else CompositeHeader(13, header_type))))))))))))))


decoder: Decoder_1[CompositeHeader] = object(_arrow2770)

__all__ = ["encoder", "decoder"]




from __future__ import annotations
from typing import Any
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.util import (equals, ignore, IEnumerable_1)
from ..Cells.fs_cell import FsCell
from ..Cells.fs_cells_collection import FsCellsCollection
from ..fs_address import FsAddress
from .fs_range_address import FsRangeAddress

def _expr197() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRangeBase", None, FsRangeBase)


class FsRangeBase:
    def __init__(self, range_address: FsRangeAddress) -> None:
        self._sortRows: Any = None
        self._sortColumns: Any = None
        self._rangeAddress: FsRangeAddress = range_address
        _id: int
        FsRangeBase.IdCounter = (FsRangeBase.IdCounter + 1) or 0
        _id = FsRangeBase.IdCounter

    def Extend(self, address: FsAddress) -> None:
        this: FsRangeBase = self
        this._rangeAddress.Extend(address)

    @property
    def RangeAddress(self, __unit: None=None) -> FsRangeAddress:
        this: FsRangeBase = self
        return this._rangeAddress

    @RangeAddress.setter
    def RangeAddress(self, range_adress: FsRangeAddress) -> None:
        this: FsRangeBase = self
        if not equals(range_adress, this._rangeAddress):
            old_address: FsRangeAddress = this._rangeAddress
            this._rangeAddress = range_adress


    def Cell(self, cell_address_in_range: FsAddress, cells: FsCellsCollection) -> FsCell:
        this: FsRangeBase = self
        abs_row: int = ((cell_address_in_range.RowNumber + this.RangeAddress.FirstAddress.RowNumber) - 1) or 0
        abs_column: int = ((cell_address_in_range.ColumnNumber + this.RangeAddress.FirstAddress.ColumnNumber) - 1) or 0
        if True if (abs_row <= 0) else (abs_row > 1048576):
            arg: int = cells.MaxRowNumber or 0
            to_fail(printf("Row number must be between 1 and %i"))(arg)

        if True if (abs_column <= 0) else (abs_column > 16384):
            arg_1: int = cells.MaxColumnNumber or 0
            to_fail(printf("Column number must be between 1 and %i"))(arg_1)

        cell: FsCell | None = cells.TryGetCell(abs_row, abs_column)
        if cell is None:
            absolute_address: FsAddress = FsAddress(abs_row, abs_column, cell_address_in_range.FixedRow, cell_address_in_range.FixedColumn)
            new_cell: FsCell = FsCell.create_empty_with_adress(absolute_address)
            this.Extend(absolute_address)
            value: None = cells.Add(new_cell, abs_row, abs_column)
            ignore(None)
            return new_cell

        else: 
            return cell


    def Cells(self, cells: FsCellsCollection) -> IEnumerable_1[FsCell]:
        this: FsRangeBase = self
        return cells.GetCellsInStringRange(this.RangeAddress.FirstAddress, this.RangeAddress.LastAddress)

    def ColumnCount(self, __unit: None=None) -> int:
        this: FsRangeBase = self
        return (this._rangeAddress.LastAddress.ColumnNumber - this._rangeAddress.FirstAddress.ColumnNumber) + 1

    def RowCount(self, __unit: None=None) -> int:
        this: FsRangeBase = self
        return (this._rangeAddress.LastAddress.RowNumber - this._rangeAddress.FirstAddress.RowNumber) + 1


FsRangeBase_reflection = _expr197

def FsRangeBase__ctor_6A2513BC(range_address: FsRangeAddress) -> FsRangeBase:
    return FsRangeBase(range_address)


def FsRangeBase__cctor(__unit: None=None) -> None:
    FsRangeBase.IdCounter = 0


FsRangeBase__cctor()

__all__ = ["FsRangeBase_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1831() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Study.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Study", string_type), ("ArcStudy", string_type), ("identifier", string_type), ("title", string_type), ("description", string_type), ("submission_date", string_type), ("public_release_date", string_type), ("publications", string_type), ("people", string_type), ("assays", string_type), ("filename", string_type), ("comments", string_type), ("protocols", string_type), ("materials", string_type), ("other_materials", string_type), ("sources", string_type), ("samples", string_type), ("process_sequence", string_type), ("factors", string_type), ("characteristic_categories", string_type), ("unit_categories", string_type), ("study_design_descriptors", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Study: str
    ArcStudy: str
    identifier: str
    title: str
    description: str
    submission_date: str
    public_release_date: str
    publications: str
    people: str
    assays: str
    filename: str
    comments: str
    protocols: str
    materials: str
    other_materials: str
    sources: str
    samples: str
    process_sequence: str
    factors: str
    characteristic_categories: str
    unit_categories: str
    study_design_descriptors: str

IContext_reflection = _expr1831

def _arrow1849(__unit: None=None) -> IEncodable:
    class ObjectExpr1832(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1833(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:Dataset")

    class ObjectExpr1834(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:identifier")

    class ObjectExpr1835(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:headline")

    class ObjectExpr1836(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:additionalType")

    class ObjectExpr1837(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:description")

    class ObjectExpr1838(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:dateCreated")

    class ObjectExpr1839(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:datePublished")

    class ObjectExpr1840(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:citation")

    class ObjectExpr1841(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:creator")

    class ObjectExpr1842(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:hasPart")

    class ObjectExpr1843(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            return helpers_11.encode_string("sdo:hasPart")

    class ObjectExpr1844(IEncodable):
        def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
            return helpers_12.encode_string("sdo:alternateName")

    class ObjectExpr1845(IEncodable):
        def Encode(self, helpers_13: IEncoderHelpers_1[Any]) -> Any:
            return helpers_13.encode_string("sdo:comment")

    class ObjectExpr1846(IEncodable):
        def Encode(self, helpers_14: IEncoderHelpers_1[Any]) -> Any:
            return helpers_14.encode_string("sdo:about")

    class ObjectExpr1847(IEncodable):
        def Encode(self, helpers_15: IEncoderHelpers_1[Any]) -> Any:
            return helpers_15.encode_string("arc:ARC#ARC_00000037")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1832()), ("Study", ObjectExpr1833()), ("identifier", ObjectExpr1834()), ("title", ObjectExpr1835()), ("additionalType", ObjectExpr1836()), ("description", ObjectExpr1837()), ("submissionDate", ObjectExpr1838()), ("publicReleaseDate", ObjectExpr1839()), ("publications", ObjectExpr1840()), ("people", ObjectExpr1841()), ("assays", ObjectExpr1842()), ("dataFiles", ObjectExpr1843()), ("filename", ObjectExpr1844()), ("comments", ObjectExpr1845()), ("processSequence", ObjectExpr1846()), ("studyDesignDescriptors", ObjectExpr1847())])
    class ObjectExpr1848(IEncodable):
        def Encode(self, helpers_16: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_16))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_16.encode_object(arg)

    return ObjectExpr1848()


context_jsonvalue: IEncodable = _arrow1849()

__all__ = ["IContext_reflection", "context_jsonvalue"]




from __future__ import annotations
from collections.abc import Callable
from pathlib import Path
import shutil
import os
from typing import Any
from ...fable_library.array_ import map
from ...fable_library.async_builder import (singleton, Async)
from ...fable_library.seq import (is_empty, empty, map as map_1, concat, append, to_array)
from ...fable_library.string_ import (starts_with_exact, trim as trim_1, replace, substring)
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, to_enumerable)
from ...fs_spreadsheet.fs_workbook import FsWorkbook
from ...fs_spreadsheet_py.fs_extension import (FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5, FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static)
from ..cross_async import all

shutil

os

Path

def directory_exists_async(path: str) -> Async[bool]:
    def _arrow3994(__unit: None=None, path: Any=path) -> Async[bool]:
        return singleton.Return(Path(path).is_dir())

    return singleton.Delay(_arrow3994)


def create_directory_async(path: str) -> Async[None]:
    def _arrow3995(__unit: None=None, path: Any=path) -> Async[None]:
        Path(path).mkdir(parents=True, exist_ok=True)
        return singleton.Zero()

    return singleton.Delay(_arrow3995)


def ensure_directory_async(path: str) -> Async[None]:
    def _arrow3997(__unit: None=None, path: Any=path) -> Async[None]:
        def _arrow3996(_arg: bool) -> Async[None]:
            return singleton.ReturnFrom(create_directory_async(path)) if (not _arg) else singleton.Zero()

        return singleton.Bind(directory_exists_async(path), _arrow3996)

    return singleton.Delay(_arrow3997)


def ensure_directory_of_file_async(file_path: str) -> Async[None]:
    def _arrow3998(__unit: None=None, file_path: Any=file_path) -> Async[None]:
        return singleton.ReturnFrom(ensure_directory_async(Path(file_path).parent))

    return singleton.Delay(_arrow3998)


def file_exists_async(path: str) -> Async[bool]:
    def _arrow3999(__unit: None=None, path: Any=path) -> Async[bool]:
        return singleton.Return(Path(path).is_file())

    return singleton.Delay(_arrow3999)


def read_file_text_async(path: str) -> Async[str]:
    def _arrow4001(__unit: None=None, path: Any=path) -> Async[str]:
        def _arrow4000(__unit: None=None) -> str:
            with open(path, 'r', encoding='utf-8') as f: return f.read()

        return singleton.Return(_arrow4000())

    return singleton.Delay(_arrow4001)


def read_file_binary_async(path: str) -> Async[bytearray]:
    def _arrow4003(__unit: None=None, path: Any=path) -> Async[bytearray]:
        def _arrow4002(__unit: None=None) -> bytearray:
            with open(path, 'rb') as f: return f.read()

        return singleton.Return(_arrow4002())

    return singleton.Delay(_arrow4003)


def read_file_xlsx_async(path: str) -> Async[FsWorkbook]:
    def _arrow4004(__unit: None=None, path: Any=path) -> Async[FsWorkbook]:
        return singleton.Return(FsSpreadsheet_FsWorkbook__FsWorkbook_fromXlsxFile_Static_Z721C83C5(path))

    return singleton.Delay(_arrow4004)


def move_file_async(old_path: str, new_path: str) -> Async[None]:
    def _arrow4005(__unit: None=None, old_path: Any=old_path, new_path: Any=new_path) -> Async[None]:
        shutil.move(old_path, new_path)
        return singleton.Zero()

    return singleton.Delay(_arrow4005)


def move_directory_async(old_path: str, new_path: str) -> Async[None]:
    return move_file_async(old_path, new_path)


def delete_file_async(path: str) -> Async[None]:
    def _arrow4006(__unit: None=None, path: Any=path) -> Async[None]:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass
        return singleton.Zero()

    return singleton.Delay(_arrow4006)


def delete_directory_async(path: str) -> Async[None]:
    def _arrow4007(__unit: None=None, path: Any=path) -> Async[None]:
        shutil.rmtree(path, ignore_errors=True)
        return singleton.Zero()

    return singleton.Delay(_arrow4007)


def write_file_text_async(path: str, text: str) -> Async[None]:
    def _arrow4008(__unit: None=None, path: Any=path, text: Any=text) -> Async[None]:
        with open(path, 'w') as f: f.write(text)
        return singleton.Zero()

    return singleton.Delay(_arrow4008)


def write_file_binary_async(path: str, bytes: bytearray) -> Async[None]:
    def _arrow4009(__unit: None=None, path: Any=path, bytes: Any=bytes) -> Async[None]:
        with open(path, 'wb') as f: f.write(bytes)
        return singleton.Zero()

    return singleton.Delay(_arrow4009)


def write_file_xlsx_async(path: str, wb: FsWorkbook) -> Async[None]:
    def _arrow4010(__unit: None=None, path: Any=path, wb: Any=wb) -> Async[None]:
        FsSpreadsheet_FsWorkbook__FsWorkbook_toXlsxFile_Static(path, wb)
        return singleton.Zero()

    return singleton.Delay(_arrow4010)


def trim(path: str) -> str:
    if starts_with_exact(path, "./"):
        return trim_1(replace(path, "./", ""), "/")

    else: 
        return trim_1(path, "/")



def make_relative(directory_path: str, path: str) -> str:
    if True if (True if (directory_path == ".") else (directory_path == "/")) else (directory_path == ""):
        return path

    else: 
        directory_path_1: str = trim(directory_path)
        path_1: str = trim(path)
        if starts_with_exact(path_1, directory_path_1):
            return substring(path_1, len(directory_path_1))

        else: 
            return path_1




def standardize_slashes(path: str) -> str:
    return replace(path, "\\", "/")


def get_sub_directories_async(path: str) -> Async[Array[str]]:
    def _arrow4011(__unit: None=None, path: Any=path) -> Async[Array[str]]:
        paths: Array[str] = [str(entry) for entry in Path(path).iterdir() if entry.is_dir()]
        return singleton.Return(map(standardize_slashes, paths, None))

    return singleton.Delay(_arrow4011)


def get_sub_files_async(path: str) -> Async[Array[str]]:
    def _arrow4012(__unit: None=None, path: Any=path) -> Async[Array[str]]:
        paths: Array[str] = [str(entry) for entry in Path(path).iterdir() if entry.is_file()]
        return singleton.Return(map(standardize_slashes, paths, None))

    return singleton.Delay(_arrow4012)


def get_all_file_paths_async(directory_path: str) -> Async[Array[str]]:
    directory_path_1: str = standardize_slashes(directory_path)
    def _arrow4018(__unit: None=None, directory_path: Any=directory_path) -> Async[Array[str]]:
        def all_files(dirs: IEnumerable_1[str]) -> Async[IEnumerable_1[str]]:
            def _arrow4016(__unit: None=None, dirs: Any=dirs) -> Async[IEnumerable_1[str]]:
                def _arrow4015(_arg: Array[Array[str]]) -> Async[IEnumerable_1[str]]:
                    sub_files_1: IEnumerable_1[str] = concat(_arg)
                    def _arrow4014(_arg_1: Array[Array[str]]) -> Async[IEnumerable_1[str]]:
                        def _arrow4013(_arg_2: Array[IEnumerable_1[str]]) -> Async[IEnumerable_1[str]]:
                            sub_dir_contents_1: IEnumerable_1[str] = concat(_arg_2)
                            return singleton.Return(append(sub_dir_contents_1, sub_files_1))

                        return singleton.Bind(all(map_1(all_files, _arg_1)), _arrow4013)

                    return singleton.Bind(all(map_1(get_sub_directories_async, dirs)), _arrow4014)

                return singleton.Return(empty()) if is_empty(dirs) else singleton.Bind(all(map_1(get_sub_files_async, dirs)), _arrow4015)

            return singleton.Delay(_arrow4016)

        def _arrow4017(_arg_3: IEnumerable_1[str]) -> Async[Array[str]]:
            def mapping_1(arg_2: str) -> str:
                return standardize_slashes(make_relative(directory_path_1, arg_2))

            all_files_relative: Array[str] = map(mapping_1, to_array(_arg_3), None)
            return singleton.Return(all_files_relative)

        return singleton.Bind(all_files(to_enumerable([directory_path_1])), _arrow4017)

    return singleton.Delay(_arrow4018)


def rename_file_or_directory_async(old_path: str, new_path: str) -> Async[None]:
    def _arrow4021(__unit: None=None, old_path: Any=old_path, new_path: Any=new_path) -> Async[None]:
        def _arrow4020(_arg: bool) -> Async[None]:
            def _arrow4019(_arg_1: bool) -> Async[None]:
                if _arg:
                    return singleton.ReturnFrom(move_file_async(old_path, new_path))

                elif _arg_1:
                    return singleton.ReturnFrom(move_directory_async(old_path, new_path))

                else: 
                    return singleton.Zero()


            return singleton.Bind(directory_exists_async(old_path), _arrow4019)

        return singleton.Bind(file_exists_async(old_path), _arrow4020)

    return singleton.Delay(_arrow4021)


def delete_file_or_directory_async(path: str) -> Async[None]:
    def _arrow4024(__unit: None=None, path: Any=path) -> Async[None]:
        def _arrow4023(_arg: bool) -> Async[None]:
            def _arrow4022(_arg_1: bool) -> Async[None]:
                if _arg:
                    return singleton.ReturnFrom(delete_file_async(path))

                elif _arg_1:
                    return singleton.ReturnFrom(delete_directory_async(path))

                else: 
                    return singleton.Zero()


            return singleton.Bind(directory_exists_async(path), _arrow4022)

        return singleton.Bind(file_exists_async(path), _arrow4023)

    return singleton.Delay(_arrow4024)


__all__ = ["directory_exists_async", "create_directory_async", "ensure_directory_async", "ensure_directory_of_file_async", "file_exists_async", "read_file_text_async", "read_file_binary_async", "read_file_xlsx_async", "move_file_async", "move_directory_async", "delete_file_async", "delete_directory_async", "write_file_text_async", "write_file_binary_async", "write_file_xlsx_async", "trim", "make_relative", "standardize_slashes", "get_sub_directories_async", "get_sub_files_async", "get_all_file_paths_async", "rename_file_or_directory_async", "delete_file_or_directory_async"]




from __future__ import annotations
from collections.abc import Callable
from datetime import datetime
from typing import Any
from ...arctrl_core.Helper.collections_ import ResizeArray_tryPick
from ...fable_library.array_ import contains
from ...fable_library.option import value as value_1
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.resize_array import exists
from ...fable_library.types import Array
from ...fable_library.util import string_hash
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .creative_work import LDCreativeWork
from .defined_term import LDDefinedTerm
from .file import LDFile
from .lab_process import LDLabProcess
from .person import LDPerson
from .property_value import LDPropertyValue
from .scholarly_article import LDScholarlyArticle
from .workflow_invocation import LDWorkflowInvocation
from .workflow_protocol import LDWorkflowProtocol

def _expr1573() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDDataset", None, LDDataset)


class LDDataset:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/Dataset"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def creator() -> str:
        return "http://schema.org/creator"

    @staticmethod
    def date_created() -> str:
        return "http://schema.org/dateCreated"

    @staticmethod
    def date_published() -> str:
        return "http://schema.org/datePublished"

    @staticmethod
    def sd_date_published() -> str:
        return "http://schema.org/datePublished"

    @staticmethod
    def license() -> str:
        return "http://schema.org/license"

    @staticmethod
    def date_modified() -> str:
        return "http://schema.org/dateModified"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def has_part() -> str:
        return "http://schema.org/hasPart"

    @staticmethod
    def headline() -> str:
        return "http://schema.org/headline"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def citation() -> str:
        return "http://schema.org/citation"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def mentions() -> str:
        return "http://schema.org/mentions"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def about() -> str:
        return "http://schema.org/about"

    @staticmethod
    def measurement_method() -> str:
        return "http://schema.org/measurementMethod"

    @staticmethod
    def measurement_technique() -> str:
        return "http://schema.org/measurementTechnique"

    @staticmethod
    def variable_measured() -> str:
        return "http://schema.org/variableMeasured"

    @staticmethod
    def main_entity() -> str:
        return "http://schema.org/mainEntity"

    @staticmethod
    def try_get_identifier_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.identifier(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_identifier_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.identifier(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("property `identifier` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `identifier` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_identifier_as_string(lp: LDNode, identifier: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.identifier(), identifier, context)

    @staticmethod
    def get_creators(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(p: LDNode, ctx: LDContext | None=None) -> bool:
            return LDPerson.validate(p, ctx)

        return lp.GetPropertyNodes(LDDataset.creator(), filter, graph, context)

    @staticmethod
    def set_creators(lp: LDNode, creators: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.creator(), creators, context)

    @staticmethod
    def try_get_date_created_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.date_created(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), datetime):
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_created_as_date_time(lp: LDNode, date_created: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.date_created(), date_created, context)

    @staticmethod
    def try_get_date_published_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.date_published(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), datetime):
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_published_as_date_time(lp: LDNode, date_published: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.date_published(), date_published, context)

    @staticmethod
    def try_get_sddate_published_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.sd_date_published(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), datetime):
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_sddate_published_as_date_time(lp: LDNode, sd_date_published: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.sd_date_published(), sd_date_published, context)

    @staticmethod
    def try_get_license_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.license(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_license_as_creative_work(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDDataset.license(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1553(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDCreativeWork.validate(n, context)

            if _arrow1553():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_license_as_string(lp: LDNode, license: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.license(), license, context)

    @staticmethod
    def set_license_as_creative_work(lp: LDNode, license: Any=None, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.license(), license, context)

    @staticmethod
    def try_get_date_modified_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.date_modified(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value_1(match_value), datetime):
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_modified_as_date_time(lp: LDNode, date_modified: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.date_modified(), date_modified, context)

    @staticmethod
    def try_get_description_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.description(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_description_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.description(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("property `description` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `description` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_description_as_string(lp: LDNode, description: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.description(), description, context)

    @staticmethod
    def get_has_parts(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDDataset.has_part(), None, graph, context)

    @staticmethod
    def get_has_parts_as_dataset(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDDataset.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.has_part(), filter, graph, context)

    @staticmethod
    def get_has_parts_as_file(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.has_part(), filter, graph, context)

    @staticmethod
    def set_has_parts(lp: LDNode, has_parts: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.has_part(), has_parts, context)

    @staticmethod
    def try_get_headline_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.headline(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.name(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("property `name` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_name_as_string(lp: LDNode, name: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.name(), name, context)

    @staticmethod
    def get_citations(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDScholarlyArticle.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.citation(), filter, graph, context)

    @staticmethod
    def set_citations(lp: LDNode, citations: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.citation(), citations, context)

    @staticmethod
    def get_comments(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.comment(), filter, graph, context)

    @staticmethod
    def set_comments(lp: LDNode, comments: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.comment(), comments, context)

    @staticmethod
    def try_get_url_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.url(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_url_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.url(), context)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                n: str = value_1(match_value)
                return n

            else: 
                raise Exception(("property `url` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `url` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_url_as_string(lp: LDNode, url: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.url(), url, context)

    @staticmethod
    def get_abouts(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDDataset.about(), None, graph, context)

    @staticmethod
    def get_abouts_as_lab_process(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDLabProcess.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.about(), filter, graph, context)

    @staticmethod
    def get_abouts_as_workflow_invocation(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDWorkflowInvocation.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.about(), filter, graph, context)

    @staticmethod
    def set_abouts(lp: LDNode, abouts: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.about(), abouts, context)

    @staticmethod
    def try_get_measurement_method_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.measurement_method(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_measurement_method_as_defined_term(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDDataset.measurement_method(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1555(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDDefinedTerm.validate(n, context)

            if _arrow1555():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_measurement_method_as_string(lp: LDNode, measurement_method: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.measurement_method(), measurement_method, context)

    @staticmethod
    def set_measurement_method_as_defined_term(lp: LDNode, measurement_method: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.measurement_method(), measurement_method, context)

    @staticmethod
    def try_get_measurement_technique_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.measurement_technique(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_measurement_technique_as_defined_term(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDDataset.measurement_technique(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1556(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDDefinedTerm.validate(n, context)

            if _arrow1556():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_measurement_technique_as_string(lp: LDNode, measurement_technique: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.measurement_technique(), measurement_technique, context)

    @staticmethod
    def set_measurement_technique_as_defined_term(lp: LDNode, measurement_technique: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.measurement_technique(), measurement_technique, context)

    @staticmethod
    def try_get_variable_measured_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDDataset.variable_measured(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value_1(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value_1(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_variable_measured_as_property_value(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = lp.TryGetPropertyAsSingleNode(LDDataset.variable_measured(), graph, context)
        (pattern_matching_result, n_1) = (None, None)
        if match_value is not None:
            def _arrow1557(__unit: None=None) -> bool:
                n: LDNode = match_value
                return LDPropertyValue.validate(n, context)

            if _arrow1557():
                pattern_matching_result = 0
                n_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_variable_measured_as_measurement_type(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        def f(n: LDNode) -> LDNode | None:
            if not LDPropertyValue.validate_fragment_descriptor(n, context):
                return n

            else: 
                return None


        return ResizeArray_tryPick(f, lp.GetPropertyNodes(LDDataset.variable_measured(), None, graph, context))

    @staticmethod
    def get_variable_measured_as_fragment_descriptors(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate_fragment_descriptor(ldnode, context_1)

        return lp.GetPropertyNodes(LDDataset.variable_measured(), filter, graph, context)

    @staticmethod
    def set_variable_measured_as_string(lp: LDNode, variable_measured: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.variable_measured(), variable_measured, context)

    @staticmethod
    def set_variable_measured_as_property_value(lp: LDNode, variable_measured: LDNode, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.variable_measured(), variable_measured, context)

    @staticmethod
    def set_variable_measured_as_property_values(lp: LDNode, variable_measured: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.variable_measured(), variable_measured, context)

    @staticmethod
    def get_main_entities(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDDataset.main_entity(), None, graph, context)

    @staticmethod
    def set_main_entities(lp: LDNode, main_entities: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDDataset.main_entity(), main_entities, context)

    @staticmethod
    def gen_idinvesigation(__unit: None=None) -> str:
        return "./"

    @staticmethod
    def gen_idstudy(identifier: str) -> str:
        return ("studies/" + identifier) + "/"

    @staticmethod
    def gen_idassay(identifier: str) -> str:
        return ("assays/" + identifier) + "/"

    @staticmethod
    def gen_idarcworkflow(identifier: str) -> str:
        return ("workflows/" + identifier) + "/"

    @staticmethod
    def gen_idarcrun(identifier: str) -> str:
        return ("runs/" + identifier) + "/"

    @staticmethod
    def validate(lp: LDNode, context: LDContext | None=None) -> bool:
        return lp.HasType(LDDataset.schema_type(), context)

    @staticmethod
    def validate_investigation(lp: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1561:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1560(x: str, y: str) -> bool:
                    return x == y

                return _arrow1560

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Investigation", lp.AdditionalType, ObjectExpr1561()) if LDDataset.validate(lp, context) else False

    @staticmethod
    def validate_study(lp: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1563:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1562(x: str, y: str) -> bool:
                    return x == y

                return _arrow1562

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Study", lp.AdditionalType, ObjectExpr1563()) if LDDataset.validate(lp, context) else False

    @staticmethod
    def validate_assay(lp: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1565:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1564(x: str, y: str) -> bool:
                    return x == y

                return _arrow1564

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Assay", lp.AdditionalType, ObjectExpr1565()) if LDDataset.validate(lp, context) else False

    @staticmethod
    def validate_arcworkflow(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> bool:
        class ObjectExpr1567:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1566(x: str, y: str) -> bool:
                    return x == y

                return _arrow1566

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        def _arrow1568(ld_0027: LDNode) -> bool:
            return LDWorkflowProtocol.validate(ld_0027, context)

        return exists(_arrow1568, LDDataset.get_main_entities(lp, graph, context)) if (contains("Workflow", lp.AdditionalType, ObjectExpr1567()) if LDDataset.validate(lp, context) else False) else False

    @staticmethod
    def validate_arcrun(lp: LDNode, context: LDContext | None=None) -> bool:
        class ObjectExpr1570:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow1569(x: str, y: str) -> bool:
                    return x == y

                return _arrow1569

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        return contains("Run", lp.AdditionalType, ObjectExpr1570()) if LDDataset.validate(lp, context) else False

    @staticmethod
    def create(id: str, identier: str | None=None, creators: Array[LDNode] | None=None, date_created: Any | None=None, date_published: Any | None=None, date_modified: Any | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, name: str | None=None, citations: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, mentions: Array[LDNode] | None=None, url: str | None=None, abouts: Array[LDNode] | None=None, measurement_method: LDNode | None=None, measurement_technique: LDNode | None=None, variable_measureds: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        s: LDNode = LDNode(id, [LDDataset.schema_type()], None, context)
        s.SetOptionalProperty(LDDataset.identifier(), identier, context)
        s.SetOptionalProperty(LDDataset.creator(), creators, context)
        s.SetOptionalProperty(LDDataset.date_created(), date_created, context)
        s.SetOptionalProperty(LDDataset.date_published(), date_published, context)
        s.SetOptionalProperty(LDDataset.date_modified(), date_modified, context)
        s.SetOptionalProperty(LDDataset.description(), description, context)
        s.SetOptionalProperty(LDDataset.has_part(), has_parts, context)
        s.SetOptionalProperty(LDDataset.name(), name, context)
        s.SetOptionalProperty(LDDataset.citation(), citations, context)
        s.SetOptionalProperty(LDDataset.comment(), comments, context)
        s.SetOptionalProperty(LDDataset.mentions(), mentions, context)
        s.SetOptionalProperty(LDDataset.url(), url, context)
        s.SetOptionalProperty(LDDataset.about(), abouts, context)
        s.SetOptionalProperty(LDDataset.measurement_method(), measurement_method, context)
        s.SetOptionalProperty(LDDataset.measurement_technique(), measurement_technique, context)
        s.SetOptionalProperty(LDDataset.variable_measured(), variable_measureds, context)
        return s

    @staticmethod
    def create_investigation(identifier: str, name: str, id: str | None=None, creators: Array[LDNode] | None=None, date_created: Any | None=None, date_published: Any | None=None, date_modified: Any | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, citations: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, mentions: Array[LDNode] | None=None, url: str | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDDataset.gen_idinvesigation() if (id is None) else id
        s: LDNode = LDDataset.create(id_1, identifier, creators, date_created, date_published, date_modified, description, has_parts, name, citations, comments, mentions, url, None, None, None, None, context)
        s.AdditionalType = ["Investigation"]
        return s

    @staticmethod
    def create_study(identifier: str, id: str | None=None, creators: Array[LDNode] | None=None, date_created: Any | None=None, date_published: Any | None=None, date_modified: Any | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, name: str | None=None, citations: Array[LDNode] | None=None, variable_measureds: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, url: str | None=None, abouts: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDDataset.gen_idstudy(identifier) if (id is None) else id
        s: LDNode = LDDataset.create(id_1, identifier, creators, date_created, date_published, date_modified, description, has_parts, name, citations, comments, None, url, abouts, None, None, variable_measureds, context)
        s.AdditionalType = ["Study"]
        return s

    @staticmethod
    def create_assay(identifier: str, id: str | None=None, name: str | None=None, description: str | None=None, creators: Array[LDNode] | None=None, has_parts: Array[LDNode] | None=None, measurement_method: LDNode | None=None, measurement_technique: LDNode | None=None, variable_measureds: Array[LDNode] | None=None, abouts: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDDataset.gen_idassay(identifier) if (id is None) else id
        s: LDNode = LDDataset.create(id_1, identifier, creators, None, None, None, description, has_parts, name, None, comments, None, None, abouts, measurement_method, measurement_technique, variable_measureds, context)
        s.AdditionalType = ["Assay"]
        return s

    @staticmethod
    def create_arcworkflow(identifier: str, main_entities: Array[LDNode], id: str | None=None, name: str | None=None, description: str | None=None, creators: Array[LDNode] | None=None, has_parts: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDDataset.gen_idarcworkflow(identifier) if (id is None) else id
        s: LDNode = LDDataset.create(id_1, identifier, creators, None, None, None, description, has_parts, name, None, comments, None, None, None, None, None, None, context)
        s.AdditionalType = ["Workflow"]
        s.SetProperty(LDDataset.main_entity(), main_entities, context)
        return s

    @staticmethod
    def create_arcrun(identifier: str, id: str | None=None, name: str | None=None, description: str | None=None, creators: Array[LDNode] | None=None, has_parts: Array[LDNode] | None=None, measurement_method: LDNode | None=None, measurement_technique: LDNode | None=None, variable_measureds: Array[LDNode] | None=None, abouts: Array[LDNode] | None=None, mentions: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        id_1: str = LDDataset.gen_idarcrun(identifier) if (id is None) else id
        s: LDNode = LDDataset.create(id_1, identifier, creators, None, None, None, description, has_parts, name, None, comments, mentions, None, abouts, measurement_method, measurement_technique, variable_measureds, context)
        s.AdditionalType = ["Run"]
        return s


LDDataset_reflection = _expr1573

__all__ = ["LDDataset_reflection"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ..arctrl_core.comment import (Remark, Comment, Comment_reflection)
from ..arctrl_core.conversion import Person_orcidKey
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_core.Table.arc_table import ArcTable
from ..arctrl_core.template import (Template, Organisation)
from ..fable_library.date import now
from ..fable_library.guid import parse
from ..fable_library.list import (of_array, FSharpList, empty, map, reverse, append as append_1, of_seq)
from ..fable_library.map_util import add_to_dict
from ..fable_library.option import to_array
from ..fable_library.reflection import (TypeInfo, class_type, string_type, list_type, record_type)
from ..fable_library.seq import (map as map_1, exists, head, append, singleton, delay, to_list, try_find, iterate_indexed, try_pick)
from ..fable_library.seq2 import List_distinct
from ..fable_library.string_ import (starts_with_exact, to_console, printf)
from ..fable_library.types import (FSharpException, Record, to_string, Array)
from ..fable_library.util import (IEnumerable_1, IEnumerator, string_hash, to_enumerable, get_enumerator, ignore)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from ..fs_spreadsheet.Tables.fs_table import FsTable
from .AnnotationTable.arc_table import (try_from_fs_worksheet, to_fs_worksheet)
from .Metadata.comment import Comment_fromString
from .Metadata.contacts import (from_rows as from_rows_1, to_rows as to_rows_1)
from .Metadata.ontology_annotation import (from_sparse_table, to_sparse_table, from_rows, to_rows)
from .Metadata.sparse_table import (SparseTable, SparseTable__TryGetValueDefault_5BAE6133, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1, SparseRowModule_tryGetValueAt, SparseRowModule_fromValues, SparseRowModule_writeToSheet, SparseRowModule_fromFsRow, SparseRowModule_getAllValues, SparseRowModule_fromAllValues)

def _expr3690() -> TypeInfo:
    return class_type("ARCtrl.Spreadsheet.TemplateReadError", None, TemplateReadError, class_type("System.Exception"))


class TemplateReadError(FSharpException):
    def __init__(self, Data0: str) -> None:
        super().__init__()
        self.Data0 = Data0


TemplateReadError_reflection = _expr3690

Metadata_ER_labels: FSharpList[str] = of_array(["ER", "ER Term Accession Number", "ER Term Source REF"])

def Metadata_ER_fromSparseTable(matrix: SparseTable) -> FSharpList[OntologyAnnotation]:
    return from_sparse_table("ER", "ER Term Source REF", "ER Term Accession Number", matrix)


def Metadata_ER_toSparseTable(designs: FSharpList[OntologyAnnotation]) -> SparseTable:
    return to_sparse_table("ER", "ER Term Source REF", "ER Term Accession Number", designs)


def Metadata_ER_fromRows(prefix: str | None, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, FSharpList[OntologyAnnotation]]:
    pattern_input: tuple[str | None, int, FSharpList[Remark], FSharpList[OntologyAnnotation]] = from_rows(prefix, "ER", "ER Term Source REF", "ER Term Accession Number", 0, rows)
    return (pattern_input[0], pattern_input[3])


def Metadata_ER_toRows(prefix: str | None, designs: FSharpList[OntologyAnnotation]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return to_rows(prefix, "ER", "ER Term Source REF", "ER Term Accession Number", designs)


Metadata_Tags_labels: FSharpList[str] = of_array(["Tags", "Tags Term Accession Number", "Tags Term Source REF"])

def Metadata_Tags_fromSparseTable(matrix: SparseTable) -> FSharpList[OntologyAnnotation]:
    return from_sparse_table("Tags", "Tags Term Source REF", "Tags Term Accession Number", matrix)


def Metadata_Tags_toSparseTable(designs: FSharpList[OntologyAnnotation]) -> SparseTable:
    return to_sparse_table("Tags", "Tags Term Source REF", "Tags Term Accession Number", designs)


def Metadata_Tags_fromRows(prefix: str | None, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, FSharpList[OntologyAnnotation]]:
    pattern_input: tuple[str | None, int, FSharpList[Remark], FSharpList[OntologyAnnotation]] = from_rows(prefix, "Tags", "Tags Term Source REF", "Tags Term Accession Number", 0, rows)
    return (pattern_input[0], pattern_input[3])


def Metadata_Tags_toRows(prefix: str | None, designs: FSharpList[OntologyAnnotation]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return to_rows(prefix, "Tags", "Tags Term Source REF", "Tags Term Accession Number", designs)


def _expr3691() -> TypeInfo:
    return record_type("ARCtrl.Spreadsheet.Metadata.Template.TemplateInfo", [], Metadata_Template_TemplateInfo, lambda: [("Id", string_type), ("Name", string_type), ("Version", string_type), ("Description", string_type), ("Organisation", string_type), ("Table", string_type), ("Comments", list_type(Comment_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class Metadata_Template_TemplateInfo(Record):
    Id: str
    Name: str
    Version: str
    Description: str
    Organisation: str
    Table: str
    Comments: FSharpList[Comment]

Metadata_Template_TemplateInfo_reflection = _expr3691

def Metadata_Template_TemplateInfo_create(id: str, name: str, version: str, description: str, organisation: str, table: str, comments: FSharpList[Comment]) -> Metadata_Template_TemplateInfo:
    return Metadata_Template_TemplateInfo(id, name, version, description, organisation, table, comments)


def Metadata_Template_TemplateInfo_get_empty(__unit: None=None) -> Metadata_Template_TemplateInfo:
    return Metadata_Template_TemplateInfo_create("", "", "", "", "", "", empty())


def Metadata_Template_TemplateInfo_get_Labels(__unit: None=None) -> FSharpList[str]:
    return of_array(["Id", "Name", "Version", "Description", "Organisation", "Table"])


def Metadata_Template_TemplateInfo_FromSparseTable_3ECCA699(matrix: SparseTable) -> Metadata_Template_TemplateInfo:
    def mapping(k: str, matrix: Any=matrix) -> Comment:
        return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, 0)))

    comments: FSharpList[Comment] = map(mapping, matrix.CommentKeys)
    return Metadata_Template_TemplateInfo_create(SparseTable__TryGetValueDefault_5BAE6133(matrix, create_missing_identifier(), ("Id", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Name", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Version", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Description", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Organisation", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Table", 0)), comments)


def Metadata_Template_TemplateInfo_ToSparseTable_48C39BE1(template: Template) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, Metadata_Template_TemplateInfo_get_Labels(), None, 2)
    comment_keys: FSharpList[str] = empty()
    def _arrow3692(__unit: None=None, template: Any=template) -> str:
        copy_of_struct: str = template.Id
        return str(copy_of_struct)

    def _arrow3693(__unit: None=None, template: Any=template) -> str:
        copy_of_struct_1: str = template.Id
        return str(copy_of_struct_1)

    add_to_dict(matrix.Matrix, ("Id", 1), "" if starts_with_exact(_arrow3692(), "MISSING_IDENTIFIER_") else _arrow3693())
    add_to_dict(matrix.Matrix, ("Name", 1), template.Name)
    add_to_dict(matrix.Matrix, ("Version", 1), template.Version)
    add_to_dict(matrix.Matrix, ("Description", 1), template.Description)
    add_to_dict(matrix.Matrix, ("Organisation", 1), to_string(template.Organisation))
    add_to_dict(matrix.Matrix, ("Table", 1), template.Table.Name)
    class ObjectExpr3695:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3694(x: str, y: str) -> bool:
                return x == y

            return _arrow3694

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3695())), matrix.ColumnCount)


def Metadata_Template_TemplateInfo_fromRows_9F97F2A(rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, Metadata_Template_TemplateInfo]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, Metadata_Template_TemplateInfo_get_Labels(), 0)
    return (tupled_arg[0], Metadata_Template_TemplateInfo_FromSparseTable_3ECCA699(tupled_arg[3]))


def Metadata_Template_TemplateInfo_toRows_48C39BE1(template: Template) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(Metadata_Template_TemplateInfo_ToSparseTable_48C39BE1(template))


def Metadata_Template_mapDeprecatedKeys(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def mapping_1(r: IEnumerable_1[tuple[int, str]], rows: Any=rows) -> IEnumerable_1[tuple[int, str]]:
        def mapping(tupled_arg: tuple[int, str], r: Any=r) -> tuple[int, str]:
            k: int = tupled_arg[0] or 0
            v: str = tupled_arg[1]
            if k == 0:
                if v == "#AUTHORS list":
                    return (k, "AUTHORS")

                elif v == "#ER list":
                    return (k, "ERS")

                elif v == "#TAGS list":
                    return (k, "TAGS")

                elif v == "Authors ORCID":
                    return (k, ("Comment[" + Person_orcidKey) + "]")

                elif v == "Authors Last Name":
                    return (k, "Author Last Name")

                elif v == "Authors First Name":
                    return (k, "Author First Name")

                elif v == "Authors Mid Initials":
                    return (k, "Author Mid Initials")

                elif v == "Authors Email":
                    return (k, "Author Email")

                elif v == "Authors Phone":
                    return (k, "Author Phone")

                elif v == "Authors Fax":
                    return (k, "Author Fax")

                elif v == "Authors Address":
                    return (k, "Author Address")

                elif v == "Authors Affiliation":
                    return (k, "Author Affiliation")

                elif v == "Authors Role":
                    return (k, "Author Roles")

                elif v == "Authors Role Term Accession Number":
                    return (k, "Author Roles Term Accession Number")

                elif v == "Authors Role Term Source REF":
                    return (k, "Author Roles Term Source REF")

                else: 
                    return (k, v)


            else: 
                return (k, v)


        return map_1(mapping, r)

    s: IEnumerable_1[IEnumerable_1[tuple[int, str]]] = map_1(mapping_1, rows)
    def predicate(v_32: str, rows: Any=rows) -> bool:
        return v_32 == "TEMPLATE"

    if exists(predicate, to_array(SparseRowModule_tryGetValueAt(0, head(s)))):
        return s

    else: 
        return append(singleton(SparseRowModule_fromValues(to_enumerable(["TEMPLATE"]))), s)



def Metadata_Template_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> tuple[Metadata_Template_TemplateInfo, FSharpList[OntologyAnnotation], FSharpList[OntologyAnnotation], FSharpList[Person]]:
    def loop(en_mut: IEnumerator[IEnumerable_1[tuple[int, str]]], last_line_mut: str | None, template_info_mut: Metadata_Template_TemplateInfo, ers_mut: FSharpList[OntologyAnnotation], tags_mut: FSharpList[OntologyAnnotation], authors_mut: FSharpList[Person], rows: Any=rows) -> tuple[Metadata_Template_TemplateInfo, FSharpList[OntologyAnnotation], FSharpList[OntologyAnnotation], FSharpList[Person]]:
        while True:
            (en, last_line, template_info, ers, tags, authors) = (en_mut, last_line_mut, template_info_mut, ers_mut, tags_mut, authors_mut)
            (pattern_matching_result,) = (None,)
            if last_line is not None:
                if last_line == "ERS":
                    pattern_matching_result = 0

                elif last_line == "TAGS":
                    pattern_matching_result = 1

                elif last_line == "AUTHORS":
                    pattern_matching_result = 2

                else: 
                    pattern_matching_result = 3


            else: 
                pattern_matching_result = 3

            if pattern_matching_result == 0:
                pattern_input: tuple[str | None, FSharpList[OntologyAnnotation]] = Metadata_ER_fromRows(None, en)
                en_mut = en
                last_line_mut = pattern_input[0]
                template_info_mut = template_info
                ers_mut = append_1(ers, pattern_input[1])
                tags_mut = tags
                authors_mut = authors
                continue

            elif pattern_matching_result == 1:
                pattern_input_1: tuple[str | None, FSharpList[OntologyAnnotation]] = Metadata_Tags_fromRows(None, en)
                en_mut = en
                last_line_mut = pattern_input_1[0]
                template_info_mut = template_info
                ers_mut = ers
                tags_mut = append_1(tags, pattern_input_1[1])
                authors_mut = authors
                continue

            elif pattern_matching_result == 2:
                pattern_input_2: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_1("Author", 0, en)
                en_mut = en
                last_line_mut = pattern_input_2[0]
                template_info_mut = template_info
                ers_mut = ers
                tags_mut = tags
                authors_mut = append_1(authors, pattern_input_2[3])
                continue

            elif pattern_matching_result == 3:
                return (template_info, ers, tags, authors)

            break

    en_1: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(Metadata_Template_mapDeprecatedKeys(rows))
    ignore(en_1.System_Collections_IEnumerator_MoveNext())
    pattern_input_3: tuple[str | None, Metadata_Template_TemplateInfo] = Metadata_Template_TemplateInfo_fromRows_9F97F2A(en_1)
    return loop(en_1, pattern_input_3[0], pattern_input_3[1], empty(), empty(), empty())


def Metadata_Template_toRows(template: Template) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def _arrow3703(__unit: None=None, template: Any=template) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3702(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3701(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3700(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    def _arrow3699(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                        def _arrow3698(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                            def _arrow3697(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                def _arrow3696(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                    return to_rows_1("Author", of_seq(template.Authors))

                                return append(singleton(SparseRowModule_fromValues(to_enumerable(["AUTHORS"]))), delay(_arrow3696))

                            return append(Metadata_Tags_toRows(None, to_list(template.Tags)), delay(_arrow3697))

                        return append(singleton(SparseRowModule_fromValues(to_enumerable(["TAGS"]))), delay(_arrow3698))

                    return append(Metadata_ER_toRows(None, to_list(template.EndpointRepositories)), delay(_arrow3699))

                return append(singleton(SparseRowModule_fromValues(to_enumerable(["ERS"]))), delay(_arrow3700))

            return append(Metadata_Template_TemplateInfo_toRows_48C39BE1(template), delay(_arrow3701))

        return append(singleton(SparseRowModule_fromValues(to_enumerable(["TEMPLATE"]))), delay(_arrow3702))

    return delay(_arrow3703)


Template_metadataSheetName: str = "isa_template"

Template_obsoleteMetadataSheetName: str = "SwateTemplateMetadata"

def Template_fromParts(template_info: Metadata_Template_TemplateInfo, ers: FSharpList[OntologyAnnotation], tags: FSharpList[OntologyAnnotation], authors: FSharpList[Person], table: ArcTable, last_updated: Any) -> Template:
    id: str = parse(template_info.Id)
    organisation: Organisation = Organisation.of_string(template_info.Organisation)
    authors_1: Array[Person] = list(authors)
    repos: Array[OntologyAnnotation] = list(ers)
    tags_1: Array[OntologyAnnotation] = list(tags)
    return Template.make(id, table, template_info.Name, template_info.Description, organisation, template_info.Version, authors_1, repos, tags_1, last_updated)


def Template_isMetadataSheetName(name: str) -> bool:
    if name == Template_metadataSheetName:
        return True

    else: 
        return name == Template_obsoleteMetadataSheetName



def Template_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return Template_isMetadataSheetName(sheet.Name)


def Template_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return Template_isMetadataSheet(sheet)

    return try_find(predicate, doc.GetWorksheets())


def Template_toMetadataSheet(template: Template) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet(Template_metadataSheetName)
    def action(row_i: int, r: IEnumerable_1[tuple[int, str]], template: Any=template) -> None:
        SparseRowModule_writeToSheet(row_i + 1, r, sheet)

    iterate_indexed(action, Metadata_Template_toRows(template))
    return sheet


def Template_fromMetadataSheet(sheet: FsWorksheet) -> tuple[Metadata_Template_TemplateInfo, FSharpList[OntologyAnnotation], FSharpList[OntologyAnnotation], FSharpList[Person]]:
    def mapping(r: FsRow, sheet: Any=sheet) -> IEnumerable_1[tuple[int, str]]:
        return SparseRowModule_fromFsRow(r)

    return Metadata_Template_fromRows(map_1(mapping, sheet.Rows))


def Template_toMetadataCollection(template: Template) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], template: Any=template) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    return map_1(mapping, Metadata_Template_toRows(template))


def Template_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> tuple[Metadata_Template_TemplateInfo, FSharpList[OntologyAnnotation], FSharpList[OntologyAnnotation], FSharpList[Person]]:
    def mapping(v: IEnumerable_1[str | None], collection: Any=collection) -> IEnumerable_1[tuple[int, str]]:
        return SparseRowModule_fromAllValues(v)

    return Metadata_Template_fromRows(map_1(mapping, collection))


def Template_fromFsWorkbook(doc: FsWorkbook) -> Template:
    pattern_input: tuple[Metadata_Template_TemplateInfo, FSharpList[OntologyAnnotation], FSharpList[OntologyAnnotation], FSharpList[Person]]
    match_value: FsWorksheet | None = Template_tryGetMetadataSheet(doc)
    if match_value is None:
        to_console(printf("Could not find metadata sheet with sheetname \"isa_template\" or deprecated sheetname \"Template\""))
        pattern_input = (Metadata_Template_TemplateInfo_get_empty(), empty(), empty(), empty())

    else: 
        pattern_input = Template_fromMetadataSheet(match_value)

    template_info: Metadata_Template_TemplateInfo = pattern_input[0]
    sheets: Array[FsWorksheet] = doc.GetWorksheets()
    def _arrow3704(__unit: None=None, doc: Any=doc) -> ArcTable:
        def try_table_name_matches(ws: FsWorksheet) -> FsWorksheet | None:
            def predicate(t: FsTable, ws: Any=ws) -> bool:
                return t.Name == template_info.Table

            if exists(predicate, ws.Tables):
                return ws

            else: 
                return None


        match_value_1: FsWorksheet | None = try_pick(try_table_name_matches, sheets)
        if match_value_1 is None:
            def try_wsname_matches(ws_1: FsWorksheet) -> FsWorksheet | None:
                if ws_1.Name == template_info.Table:
                    return ws_1

                else: 
                    return None


            match_value_3: FsWorksheet | None = try_pick(try_wsname_matches, sheets)
            if match_value_3 is None:
                raise TemplateReadError(("No worksheet or table with name `" + template_info.Table) + "` found")

            else: 
                ws_3: FsWorksheet = match_value_3
                match_value_4: ArcTable | None = try_from_fs_worksheet(ws_3)
                if match_value_4 is None:
                    raise TemplateReadError(("Ws with name `" + ws_3.Name) + "` could not be converted to a table")

                else: 
                    return match_value_4



        else: 
            ws_2: FsWorksheet = match_value_1
            match_value_2: ArcTable | None = try_from_fs_worksheet(ws_2)
            if match_value_2 is None:
                raise TemplateReadError(("Ws with name `" + ws_2.Name) + "` could not be converted to a table")

            else: 
                return match_value_2



    return Template_fromParts(template_info, pattern_input[1], pattern_input[2], pattern_input[3], _arrow3704(), now())


def Template_toFsWorkbook(template: Template) -> FsWorkbook:
    doc: FsWorkbook = FsWorkbook()
    meta_data_sheet: FsWorksheet = Template_toMetadataSheet(template)
    doc.AddWorksheet(meta_data_sheet)
    sheet: FsWorksheet = to_fs_worksheet(None, template.Table)
    doc.AddWorksheet(sheet)
    return doc


__all__ = ["TemplateReadError_reflection", "Metadata_ER_labels", "Metadata_ER_fromSparseTable", "Metadata_ER_toSparseTable", "Metadata_ER_fromRows", "Metadata_ER_toRows", "Metadata_Tags_labels", "Metadata_Tags_fromSparseTable", "Metadata_Tags_toSparseTable", "Metadata_Tags_fromRows", "Metadata_Tags_toRows", "Metadata_Template_TemplateInfo_reflection", "Metadata_Template_TemplateInfo_create", "Metadata_Template_TemplateInfo_get_empty", "Metadata_Template_TemplateInfo_get_Labels", "Metadata_Template_TemplateInfo_FromSparseTable_3ECCA699", "Metadata_Template_TemplateInfo_ToSparseTable_48C39BE1", "Metadata_Template_TemplateInfo_fromRows_9F97F2A", "Metadata_Template_TemplateInfo_toRows_48C39BE1", "Metadata_Template_mapDeprecatedKeys", "Metadata_Template_fromRows", "Metadata_Template_toRows", "Template_metadataSheetName", "Template_obsoleteMetadataSheetName", "Template_fromParts", "Template_isMetadataSheetName", "Template_isMetadataSheet", "Template_tryGetMetadataSheet", "Template_toMetadataSheet", "Template_fromMetadataSheet", "Template_toMetadataCollection", "Template_fromMetadataCollection", "Template_fromFsWorkbook", "Template_toFsWorkbook"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.comment import Comment
from ..arctrl_core.comment_list import try_item
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.Process.component import (Component, Component_fromISAString_7C9A7CF8)
from ..arctrl_core.Process.factor import Factor
from ..arctrl_core.Process.factor_value import FactorValue
from ..arctrl_core.Process.material_attribute import (MaterialAttribute, MaterialAttribute_fromString_5980DC03)
from ..arctrl_core.Process.material_attribute_value import MaterialAttributeValue
from ..arctrl_core.Process.process_parameter_value import ProcessParameterValue
from ..arctrl_core.Process.protocol_parameter import ProtocolParameter
from ..fable_library.int32 import try_parse
from ..fable_library.option import (bind, value)
from ..fable_library.seq import find_index
from ..fable_library.types import (FSharpRef, Array)
from ..fable_library.util import (int32_to_string, equals)

def try_int(str_1: str) -> int | None:
    match_value: tuple[bool, int]
    out_arg: int = 0
    def _arrow3778(__unit: None=None, str_1: Any=str_1) -> int:
        return out_arg

    def _arrow3779(v: int, str_1: Any=str_1) -> None:
        nonlocal out_arg
        out_arg = v or 0

    match_value = (try_parse(str_1, 511, False, 32, FSharpRef(_arrow3778, _arrow3779)), out_arg)
    if match_value[0]:
        return match_value[1]

    else: 
        return None



order_name: str = "ColumnIndex"

def create_order_comment(index: int) -> Comment:
    return Comment.create(order_name, int32_to_string(index))


def try_get_index(comments: Array[Comment]) -> int | None:
    match_value: str | None = try_item(order_name, comments)
    if match_value is not None:
        ci: str = match_value
        def predicate(c: Comment, comments: Any=comments) -> bool:
            return equals(c.Name, order_name)

        i: int = find_index(predicate, comments) or 0
        comments.pop(i)
        return try_int(ci)

    else: 
        return None



def set_ontology_annotation_index_inplace(i: int, oa: OntologyAnnotation) -> None:
    (oa.Comments.append(create_order_comment(i)))


def set_ontology_annotation_index(i: int, oa: OntologyAnnotation) -> OntologyAnnotation:
    oac: OntologyAnnotation = oa.Copy()
    set_ontology_annotation_index_inplace(i, oac)
    return oac


def try_get_ontology_annotation_index(oa: OntologyAnnotation) -> int | None:
    return try_get_index(oa.Comments)


def try_get_parameter_index(param: ProtocolParameter) -> int | None:
    def binder(oa: OntologyAnnotation, param: Any=param) -> int | None:
        return try_get_index(oa.Comments)

    return bind(binder, param.ParameterName)


def try_get_parameter_column_index(param_value: ProcessParameterValue) -> int | None:
    def _arrow3780(param: ProtocolParameter, param_value: Any=param_value) -> int | None:
        return try_get_parameter_index(param)

    return bind(_arrow3780, param_value.Category)


def try_get_factor_index(factor: Factor) -> int | None:
    def binder(oa: OntologyAnnotation, factor: Any=factor) -> int | None:
        return try_get_index(oa.Comments)

    return bind(binder, factor.FactorType)


def try_get_factor_column_index(factor_value: FactorValue) -> int | None:
    def _arrow3781(factor: Factor, factor_value: Any=factor_value) -> int | None:
        return try_get_factor_index(factor)

    return bind(_arrow3781, factor_value.Category)


def try_get_characteristic_index(characteristic: MaterialAttribute) -> int | None:
    def binder(oa: OntologyAnnotation, characteristic: Any=characteristic) -> int | None:
        return try_get_index(oa.Comments)

    return bind(binder, characteristic.CharacteristicType)


def try_get_characteristic_column_index(characteristic_value: MaterialAttributeValue) -> int | None:
    def _arrow3782(characteristic: MaterialAttribute, characteristic_value: Any=characteristic_value) -> int | None:
        return try_get_characteristic_index(characteristic)

    return bind(_arrow3782, characteristic_value.Category)


def try_get_component_index(comp: Component) -> int | None:
    def binder(oa: OntologyAnnotation, comp: Any=comp) -> int | None:
        return try_get_index(oa.Comments)

    return bind(binder, comp.ComponentType)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_fromStringWithColumnIndex_Static(name: str, term: str, source: str, accession: str, value_index: int) -> Factor:
    return Factor.from_string(name, term, source, accession, [create_order_comment(value_index)])


def ARCtrl_OntologyAnnotation__OntologyAnnotation_getColumnIndex_Static_ZDED3A0F(f: OntologyAnnotation) -> int:
    return value(try_get_ontology_annotation_index(f))


def ARCtrl_OntologyAnnotation__OntologyAnnotation_GetColumnIndex(this: OntologyAnnotation) -> int:
    return value(try_get_ontology_annotation_index(this))


def ARCtrl_OntologyAnnotation__OntologyAnnotation_tryGetColumnIndex_Static_ZDED3A0F(f: OntologyAnnotation) -> int | None:
    return try_get_ontology_annotation_index(f)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_TryGetColumnIndex(this: OntologyAnnotation) -> int | None:
    return try_get_ontology_annotation_index(this)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_setColumnIndex_Static(i: int, oa: OntologyAnnotation) -> OntologyAnnotation:
    return set_ontology_annotation_index(i, oa)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_SetColumnIndex_Z524259A4(this: OntologyAnnotation, i: int) -> None:
    set_ontology_annotation_index_inplace(i, this)


def ARCtrl_Process_Factor__Factor_fromStringWithColumnIndex_Static(name: str, term: str, source: str, accession: str, value_index: int) -> Factor:
    return Factor.from_string(name, term, source, accession, [create_order_comment(value_index)])


def ARCtrl_Process_Factor__Factor_getColumnIndex_Static_7206F0D9(f: Factor) -> int:
    return value(try_get_factor_index(f))


def ARCtrl_Process_Factor__Factor_GetColumnIndex(this: Factor) -> int:
    return value(try_get_factor_index(this))


def ARCtrl_Process_Factor__Factor_tryGetColumnIndex_Static_7206F0D9(f: Factor) -> int | None:
    return try_get_factor_index(f)


def ARCtrl_Process_Factor__Factor_TryGetColumnIndex(this: Factor) -> int | None:
    return try_get_factor_index(this)


def ARCtrl_Process_FactorValue__FactorValue_getColumnIndex_Static_7105C732(f: FactorValue) -> int:
    return value(try_get_factor_column_index(f))


def ARCtrl_Process_FactorValue__FactorValue_GetColumnIndex(this: FactorValue) -> int:
    return value(try_get_factor_column_index(this))


def ARCtrl_Process_FactorValue__FactorValue_tryGetColumnIndex_Static_7105C732(f: FactorValue) -> int | None:
    return try_get_factor_column_index(f)


def ARCtrl_Process_FactorValue__FactorValue_TryGetColumnIndex(this: FactorValue) -> int | None:
    return try_get_factor_column_index(this)


def ARCtrl_Process_MaterialAttribute__MaterialAttribute_fromStringWithColumnIndex_Static(term: str, source: str, accession: str, value_index: int) -> MaterialAttribute:
    return MaterialAttribute_fromString_5980DC03(term, source, accession, [create_order_comment(value_index)])


def ARCtrl_Process_MaterialAttribute__MaterialAttribute_getColumnIndex_Static_Z1E3B85DD(m: MaterialAttribute) -> int:
    return value(try_get_characteristic_index(m))


def ARCtrl_Process_MaterialAttribute__MaterialAttribute_GetColumnIndex(this: MaterialAttribute) -> int:
    return value(try_get_characteristic_index(this))


def ARCtrl_Process_MaterialAttribute__MaterialAttribute_tryGetColumnIndex_Static_Z1E3B85DD(m: MaterialAttribute) -> int | None:
    return try_get_characteristic_index(m)


def ARCtrl_Process_MaterialAttribute__MaterialAttribute_TryGetColumnIndex(this: MaterialAttribute) -> int | None:
    return try_get_characteristic_index(this)


def ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_getColumnIndex_Static_Z772273B8(m: MaterialAttributeValue) -> int:
    return value(try_get_characteristic_column_index(m))


def ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_GetColumnIndex(this: MaterialAttributeValue) -> int:
    return value(try_get_characteristic_column_index(this))


def ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_tryGetColumnIndex_Static_Z772273B8(m: MaterialAttributeValue) -> int | None:
    return try_get_characteristic_column_index(m)


def ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_TryGetColumnIndex(this: MaterialAttributeValue) -> int | None:
    return try_get_characteristic_column_index(this)


def ARCtrl_Process_ProtocolParameter__ProtocolParameter_fromStringWithColumnIndex_Static(term: str, source: str, accession: str, value_index: int) -> ProtocolParameter:
    return ProtocolParameter.from_string(term, source, accession, [create_order_comment(value_index)])


def ARCtrl_Process_ProtocolParameter__ProtocolParameter_getColumnIndex_Static_Z11F87B15(p: ProtocolParameter) -> int:
    return value(try_get_parameter_index(p))


def ARCtrl_Process_ProtocolParameter__ProtocolParameter_GetColumnIndex(this: ProtocolParameter) -> int:
    return value(try_get_parameter_index(this))


def ARCtrl_Process_ProtocolParameter__ProtocolParameter_tryGetColumnIndex_Static_Z11F87B15(p: ProtocolParameter) -> int | None:
    return try_get_parameter_index(p)


def ARCtrl_Process_ProtocolParameter__ProtocolParameter_TryGetColumnIndex(this: ProtocolParameter) -> int | None:
    return try_get_parameter_index(this)


def ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_getColumnIndex_Static_Z1576263(p: ProcessParameterValue) -> int:
    return value(try_get_parameter_column_index(p))


def ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_GetColumnIndex(this: ProcessParameterValue) -> int:
    return value(try_get_parameter_column_index(this))


def ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_tryGetColumnIndex_Static_Z1576263(p: ProcessParameterValue) -> int | None:
    return try_get_parameter_column_index(p)


def ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_TryGetColumnIndex(this: ProcessParameterValue) -> int | None:
    return try_get_parameter_column_index(this)


def ARCtrl_Process_Component__Component_fromStringWithColumnIndex_Static(name: str, term: str, source: str, accession: str, value_index: int) -> Component:
    return Component_fromISAString_7C9A7CF8(name, term, source, accession, [create_order_comment(value_index)])


def ARCtrl_Process_Component__Component_getColumnIndex_Static_Z685B8F25(f: Component) -> int:
    return value(try_get_component_index(f))


def ARCtrl_Process_Component__Component_GetColumnIndex(this: Component) -> int:
    return value(try_get_component_index(this))


def ARCtrl_Process_Component__Component_tryGetColumnIndex_Static_Z685B8F25(f: Component) -> int | None:
    return try_get_component_index(f)


def ARCtrl_Process_Component__Component_TryGetColumnIndex(this: Component) -> int | None:
    return try_get_component_index(this)


__all__ = ["try_int", "order_name", "create_order_comment", "try_get_index", "set_ontology_annotation_index_inplace", "set_ontology_annotation_index", "try_get_ontology_annotation_index", "try_get_parameter_index", "try_get_parameter_column_index", "try_get_factor_index", "try_get_factor_column_index", "try_get_characteristic_index", "try_get_characteristic_column_index", "try_get_component_index", "ARCtrl_OntologyAnnotation__OntologyAnnotation_fromStringWithColumnIndex_Static", "ARCtrl_OntologyAnnotation__OntologyAnnotation_getColumnIndex_Static_ZDED3A0F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_GetColumnIndex", "ARCtrl_OntologyAnnotation__OntologyAnnotation_tryGetColumnIndex_Static_ZDED3A0F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_TryGetColumnIndex", "ARCtrl_OntologyAnnotation__OntologyAnnotation_setColumnIndex_Static", "ARCtrl_OntologyAnnotation__OntologyAnnotation_SetColumnIndex_Z524259A4", "ARCtrl_Process_Factor__Factor_fromStringWithColumnIndex_Static", "ARCtrl_Process_Factor__Factor_getColumnIndex_Static_7206F0D9", "ARCtrl_Process_Factor__Factor_GetColumnIndex", "ARCtrl_Process_Factor__Factor_tryGetColumnIndex_Static_7206F0D9", "ARCtrl_Process_Factor__Factor_TryGetColumnIndex", "ARCtrl_Process_FactorValue__FactorValue_getColumnIndex_Static_7105C732", "ARCtrl_Process_FactorValue__FactorValue_GetColumnIndex", "ARCtrl_Process_FactorValue__FactorValue_tryGetColumnIndex_Static_7105C732", "ARCtrl_Process_FactorValue__FactorValue_TryGetColumnIndex", "ARCtrl_Process_MaterialAttribute__MaterialAttribute_fromStringWithColumnIndex_Static", "ARCtrl_Process_MaterialAttribute__MaterialAttribute_getColumnIndex_Static_Z1E3B85DD", "ARCtrl_Process_MaterialAttribute__MaterialAttribute_GetColumnIndex", "ARCtrl_Process_MaterialAttribute__MaterialAttribute_tryGetColumnIndex_Static_Z1E3B85DD", "ARCtrl_Process_MaterialAttribute__MaterialAttribute_TryGetColumnIndex", "ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_getColumnIndex_Static_Z772273B8", "ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_GetColumnIndex", "ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_tryGetColumnIndex_Static_Z772273B8", "ARCtrl_Process_MaterialAttributeValue__MaterialAttributeValue_TryGetColumnIndex", "ARCtrl_Process_ProtocolParameter__ProtocolParameter_fromStringWithColumnIndex_Static", "ARCtrl_Process_ProtocolParameter__ProtocolParameter_getColumnIndex_Static_Z11F87B15", "ARCtrl_Process_ProtocolParameter__ProtocolParameter_GetColumnIndex", "ARCtrl_Process_ProtocolParameter__ProtocolParameter_tryGetColumnIndex_Static_Z11F87B15", "ARCtrl_Process_ProtocolParameter__ProtocolParameter_TryGetColumnIndex", "ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_getColumnIndex_Static_Z1576263", "ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_GetColumnIndex", "ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_tryGetColumnIndex_Static_Z1576263", "ARCtrl_Process_ProcessParameterValue__ProcessParameterValue_TryGetColumnIndex", "ARCtrl_Process_Component__Component_fromStringWithColumnIndex_Static", "ARCtrl_Process_Component__Component_getColumnIndex_Static_Z685B8F25", "ARCtrl_Process_Component__Component_GetColumnIndex", "ARCtrl_Process_Component__Component_tryGetColumnIndex_Static_Z685B8F25", "ARCtrl_Process_Component__Component_TryGetColumnIndex"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_iter)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.factor import Factor
from ...fable_library.list import (of_array, FSharpList, length, singleton, initialize, map, empty, iterate_indexed, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import default_arg
from ...fable_library.seq2 import List_distinct
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

name_label: str = "Name"

factor_type_label: str = "Type"

type_term_accession_number_label: str = "Type Term Accession Number"

type_term_source_reflabel: str = "Type Term Source REF"

labels: FSharpList[str] = of_array([name_label, factor_type_label, type_term_accession_number_label, type_term_source_reflabel])

def from_string(name: str | None, design_type: str | None, type_term_source_ref: str | None, type_term_accession_number: str | None, comments: Array[Comment]) -> Factor:
    factor_type: OntologyAnnotation = OntologyAnnotation.create(design_type, type_term_source_ref, type_term_accession_number)
    factor_type_1: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), factor_type)
    comments_1: Array[Comment] | None = Option_fromValueWithDefault([], comments)
    return Factor.make(name, factor_type_1, comments_1)


def from_sparse_table(matrix: SparseTable) -> FSharpList[Factor]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        comments: Array[Comment] = SparseTable_GetEmptyComments_3ECCA699(matrix)
        return singleton(Factor.create(None, None, comments))

    else: 
        def _arrow3451(i: int, matrix: Any=matrix) -> Factor:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: Array[Comment] = list(map(mapping, matrix.CommentKeys))
            return from_string(SparseTable__TryGetValue_11FD62A8(matrix, (name_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (factor_type_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_source_reflabel, i)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_accession_number_label, i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3451)



def to_sparse_table(factors: FSharpList[Factor]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, length(factors) + 1)
    comment_keys: FSharpList[str] = empty()
    def action(i: int, f: Factor, factors: Any=factors) -> None:
        i_1: int = (i + 1) or 0
        ft: dict[str, Any]
        f_1: OntologyAnnotation = default_arg(f.FactorType, OntologyAnnotation())
        ft = OntologyAnnotation.to_string_object(f_1, True)
        add_to_dict(matrix.Matrix, (name_label, i_1), default_arg(f.Name, ""))
        add_to_dict(matrix.Matrix, (factor_type_label, i_1), ft["TermName"])
        add_to_dict(matrix.Matrix, (type_term_accession_number_label, i_1), ft["TermAccessionNumber"])
        add_to_dict(matrix.Matrix, (type_term_source_reflabel, i_1), ft["TermSourceREF"])
        match_value: Array[Comment] | None = f.Comments
        if match_value is not None:
            def f_2(comment: Comment, i: Any=i, f: Any=f) -> None:
                nonlocal comment_keys
                pattern_input: tuple[str, str] = Comment_toString(comment)
                n: str = pattern_input[0]
                comment_keys = cons(n, comment_keys)
                add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

            ResizeArray_iter(f_2, match_value)


    iterate_indexed(action, factors)
    class ObjectExpr3453:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3452(x: str, y: str) -> bool:
                return x == y

            return _arrow3452

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3453())), matrix.ColumnCount)


def from_rows(prefix: str | None, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[Factor]]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number) if (prefix is None) else SparseTable_FromRows_Z5579EC29(rows, labels, line_number, prefix)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(prefix: str | None, factors: FSharpList[Factor]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    m: SparseTable = to_sparse_table(factors)
    if prefix is None:
        return SparseTable_ToRows_759CAFC1(m)

    else: 
        return SparseTable_ToRows_759CAFC1(m, prefix)



__all__ = ["name_label", "factor_type_label", "type_term_accession_number_label", "type_term_source_reflabel", "labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from typing import (Any, Protocol, TypeVar)
from ..fable_library.array_ import (append, map as map_1)
from ..fable_library.map_util import add_to_dict
from ..fable_library.mutable_map import Dictionary
from ..fable_library.option import value as value_1
from ..fable_library.reg_exp import (is_match, create)
from ..fable_library.seq import (iterate, to_array, choose, map)
from ..fable_library.types import Array
from ..fable_library.util import (equals, structural_hash, IEnumerable_1, get_enumerator, dispose)
from .property_helper import PropertyHelper

__A_ = TypeVar("__A_")

_U = TypeVar("_U")

_T = TypeVar("_T")

def Dictionary_ofSeq(s: IEnumerable_1[Any]) -> Any:
    class ObjectExpr35:
        @property
        def Equals(self) -> Callable[[__A_, __A_], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[__A_], int]:
            return structural_hash

    d: Any = Dictionary([], ObjectExpr35())
    def action(kv: Any, s: Any=s) -> None:
        add_to_dict(d, kv[0], kv[1])

    iterate(action, s)
    return d


def Dictionary_choose(f: Callable[[_T], _U | None], d: Any) -> Any:
    class ObjectExpr36:
        @property
        def Equals(self) -> Callable[[__A_, __A_], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[__A_], int]:
            return structural_hash

    nd: Any = Dictionary([], ObjectExpr36())
    enumerator: Any = get_enumerator(d)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            match_value: _U | None = f(kv[1])
            if match_value is None:
                pass

            else: 
                add_to_dict(nd, kv[0], value_1(match_value))


    finally: 
        dispose(enumerator)

    return nd


class PropertyObject(Protocol):
    @property
    @abstractmethod
    def fget(self) -> Any:
        ...

    @property
    @abstractmethod
    def fset(self) -> Any:
        ...


def PropertyObjectModule_getGetter(o: PropertyObject) -> Callable[[Any], Any]:
    match_value: Callable[[Any], Any] | None = o.fget
    if match_value is None:
        def _arrow37(o_1: Any=None, o: Any=o) -> Any:
            raise Exception("Property does not contain getter")

        return _arrow37

    else: 
        return match_value



def PropertyObjectModule_getSetter(o: PropertyObject) -> Callable[[Any, Any], None]:
    match_value: Callable[[Any, Any], None] | None = o.fset
    if match_value is None:
        def _arrow39(s: Any=None, o: Any=o) -> Callable[[Any], None]:
            def _arrow38(o_1: Any=None) -> None:
                raise Exception("Property does not contain setter")

            return _arrow38

        return _arrow39

    else: 
        return match_value



def PropertyObjectModule_containsGetter(o: PropertyObject) -> bool:
    match_value: Callable[[Any], Any] | None = o.fget
    if match_value is None:
        return False

    else: 
        return True



def PropertyObjectModule_containsSetter(o: PropertyObject) -> bool:
    match_value: Callable[[Any, Any], None] | None = o.fset
    if match_value is None:
        return False

    else: 
        return True



def PropertyObjectModule_isWritable(o: PropertyObject) -> bool:
    return PropertyObjectModule_containsSetter(o)


def PropertyObjectModule_tryProperty(o: Any=None) -> PropertyObject | None:
    if isinstance(o, property):
        return o

    else: 
        return None



def create_getter(prop_name: str, o: Any=None) -> Any:
    return getattr(o,prop_name)


def create_setter(prop_name: str, o: Any=None, value: Any=None) -> None:
    setattr(o,prop_name,value)


def get_static_property_objects(o: Any=None) -> Any:
    def f(o_2: Any=None, o: Any=o) -> PropertyObject | None:
        return PropertyObjectModule_tryProperty(o_2)

    def _arrow40(__unit: None=None, o: Any=o) -> Any:
        o_1: Any = o.__class__
        return vars(o_1).items()

    return Dictionary_choose(f, _arrow40())


def remove_static_property_value(o: Any, prop_name: str) -> None:
    setattr(o,prop_name,None)


def create_remover(prop_name: str, is_static: bool) -> Callable[[Any], None]:
    if is_static:
        def _arrow41(o: Any=None, prop_name: Any=prop_name, is_static: Any=is_static) -> None:
            remove_static_property_value(o, prop_name)

        return _arrow41

    else: 
        def _arrow42(o_1: Any=None, prop_name: Any=prop_name, is_static: Any=is_static) -> None:
            delattr(o_1,prop_name)

        return _arrow42



def try_get_property_object(o: Any, prop_name: str) -> PropertyObject | None:
    match_value: PropertyObject | None = PropertyObjectModule_tryProperty(o.__dict__.get(prop_name))
    if match_value is None:
        return None

    else: 
        return match_value



def try_get_dynamic_property_helper(o: Any, prop_name: str) -> PropertyHelper | None:
    match_value: Any | None = o.__dict__.get(prop_name)
    if match_value is None:
        return None

    else: 
        def _arrow43(o_1: Any=None, o: Any=o, prop_name: Any=prop_name) -> Any:
            return create_getter(prop_name, o_1)

        def _arrow44(o_2: Any=None, value: Any=None, o: Any=o, prop_name: Any=prop_name) -> None:
            create_setter(prop_name, o_2, value)

        def _arrow45(o_3: Any=None, o: Any=o, prop_name: Any=prop_name) -> None:
            delattr(o_3,prop_name)

        return PropertyHelper(prop_name, False, True, True, False, _arrow43, _arrow44, _arrow45)



def try_get_static_property_helper(o: Any, prop_name: str) -> PropertyHelper | None:
    match_value: PropertyObject | None = try_get_property_object(o.__class__, prop_name)
    if match_value is None:
        return None

    else: 
        is_writable: bool = PropertyObjectModule_isWritable(match_value)
        def _arrow46(o_1: Any=None, o: Any=o, prop_name: Any=prop_name) -> Any:
            return create_getter(prop_name, o_1)

        def _arrow47(o_2: Any=None, value: Any=None, o: Any=o, prop_name: Any=prop_name) -> None:
            create_setter(prop_name, o_2, value)

        def _arrow48(o_3: Any=None, o: Any=o, prop_name: Any=prop_name) -> None:
            remove_static_property_value(o_3, prop_name)

        return PropertyHelper(prop_name, True, False, is_writable, not is_writable, _arrow46, _arrow47, _arrow48)



transpiled_property_regex: str = "^([a-zA-Z]+_)+[0-9]+$"

def is_transpiled_property_helper(property_name: str) -> bool:
    return is_match(create(transpiled_property_regex), property_name)


def get_dynamic_property_helpers(o: Any=None) -> Array[PropertyHelper]:
    def chooser(kv: Any, o: Any=o) -> PropertyHelper | None:
        n: str = kv[0]
        if is_transpiled_property_helper(n):
            return None

        else: 
            def _arrow49(o_1: Any=None, kv: Any=kv) -> Any:
                return create_getter(n, o_1)

            def _arrow50(o_2: Any=None, value: Any=None, kv: Any=kv) -> None:
                create_setter(n, o_2, value)

            def _arrow51(o_3: Any=None, kv: Any=kv) -> None:
                delattr(o_3,n)

            return PropertyHelper(n, False, True, True, False, _arrow49, _arrow50, _arrow51)


    return to_array(choose(chooser, vars(o).items()))


def get_static_property_helpers(o: Any=None) -> Array[PropertyHelper]:
    def mapping(kv: Any, o: Any=o) -> PropertyHelper:
        n: str = kv[0]
        po: PropertyObject = kv[1]
        def _arrow52(o_1: Any=None, kv: Any=kv) -> Any:
            return create_getter(n, o_1)

        def _arrow53(o_2: Any=None, value: Any=None, kv: Any=kv) -> None:
            create_setter(n, o_2, value)

        def _arrow54(o_3: Any=None, kv: Any=kv) -> None:
            remove_static_property_value(o_3, n)

        return PropertyHelper(n, True, False, PropertyObjectModule_isWritable(po), not PropertyObjectModule_isWritable(po), _arrow52, _arrow53, _arrow54)

    return to_array(map(mapping, get_static_property_objects(o)))


def get_property_helpers(o: Any=None) -> Array[PropertyHelper]:
    array_1: Array[PropertyHelper] = get_dynamic_property_helpers(o)
    return append(get_static_property_helpers(o), array_1, None)


def get_property_names(o: Any=None) -> Array[str]:
    def mapping(h: PropertyHelper, o: Any=o) -> str:
        return h.Name

    return map_1(mapping, get_property_helpers(o), None)


__all__ = ["Dictionary_ofSeq", "Dictionary_choose", "PropertyObjectModule_getGetter", "PropertyObjectModule_getSetter", "PropertyObjectModule_containsGetter", "PropertyObjectModule_containsSetter", "PropertyObjectModule_isWritable", "PropertyObjectModule_tryProperty", "create_getter", "create_setter", "get_static_property_objects", "remove_static_property_value", "create_remover", "try_get_property_object", "try_get_dynamic_property_helper", "try_get_static_property_helper", "transpiled_property_regex", "is_transpiled_property_helper", "get_dynamic_property_helpers", "get_static_property_helpers", "get_property_helpers", "get_property_names"]




from decimal import MAX_EMAX, MIN_EMIN, Decimal, getcontext

from .types import FSharpRef


getcontext().prec = 29

get_zero = Decimal(0)
get_one = Decimal(1)

get_minus_one = Decimal(-1)
get_max_value = MAX_EMAX
get_min_value = MIN_EMIN


def compare(x: Decimal, y: Decimal) -> int:
    return -1 if x < y else 1 if x > y else 0


def equals(a: Decimal, b: Decimal) -> bool:
    return a == b


def abs(x: Decimal) -> Decimal:
    return -x if x < 0 else x


def sign(x: Decimal) -> int:
    return -1 if x < 0 else 1 if x > 0 else 0


def max(x: Decimal, y: Decimal) -> Decimal:
    return x if x > y else y


def min(x: Decimal, y: Decimal) -> Decimal:
    return x if x < y else y


def add(a: Decimal, b: Decimal) -> Decimal:
    return a + b


def subtract(a: Decimal, b: Decimal) -> Decimal:
    return a - b


def multiply(a: Decimal, b: Decimal) -> Decimal:
    return a * b


def divide(a: Decimal, b: Decimal) -> Decimal:
    return a / b


def remainder(a: Decimal, b: Decimal) -> Decimal:
    return a % b


def negate(a: Decimal) -> Decimal:
    return -a


def op_unary_negation(a: Decimal) -> Decimal:
    return -a


def op_unary_plus(a: Decimal) -> Decimal:
    return +a


def op_addition(a: Decimal, b: Decimal) -> Decimal:
    return a + b


def op_subtraction(a: Decimal, b: Decimal) -> Decimal:
    return a - b


def op_multiply(a: Decimal, b: Decimal) -> Decimal:
    return a * b


def op_division(a: Decimal, b: Decimal) -> Decimal:
    return a / b


def op_modulus(a: Decimal, b: Decimal) -> Decimal:
    return a % b


def op_less_than(a: Decimal, b: Decimal) -> bool:
    return a < b


def op_less_than_or_equal(a: Decimal, b: Decimal) -> bool:
    return a <= b


def op_greater_than(a: Decimal, b: Decimal) -> bool:
    return a > b


def op_greater_than_or_equal(a: Decimal, b: Decimal) -> bool:
    return a >= b


def op_equality(a: Decimal, b: Decimal) -> bool:
    return a == b


def op_inequality(a: Decimal, b: Decimal) -> bool:
    return a != b


def from_parts(low: int, mid: int, high: int, isNegative: bool, scale: int) -> Decimal:
    sign = -1 if isNegative else 1

    if low < 0:
        low = 0x100000000 + low

    if mid < 0:
        mid = 0xFFFFFFFF00000000 + mid + 1
    else:
        mid = mid << 32

    if high < 0:
        high = 0xFFFFFFFF0000000000000000 + high + 1
    else:
        high = high << 64

    value = Decimal((low + mid + high) * sign)
    if scale:
        dscale = Decimal(pow(10, scale))
        return value / dscale
    return value


def to_string(x: Decimal) -> str:
    return str(x)


def to_number(x: Decimal) -> float:
    return float(x)


def parse(string: str) -> Decimal:
    return Decimal(string)


def try_parse(string: str, def_value: FSharpRef[Decimal]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


__all__ = [
    "compare",
    "equals",
    "abs",
    "sign",
    "max",
    "min",
    "add",
    "subtract",
    "multiply",
    "divide",
    "remainder",
    "negate",
    "op_unary_negation",
    "op_unary_plus",
    "op_addition",
    "op_subtraction",
    "op_multiply",
    "op_division",
    "op_modulus",
    "op_less_than",
    "op_less_than_or_equal",
    "op_greater_than",
    "op_greater_than_or_equal",
    "op_equality",
    "op_inequality",
    "from_parts",
    "to_string",
    "to_number",
    "try_parse",
    "parse",
]
